class InvalidAddressError(Exception):
    """
    The specified address is invalid.
    """

    def __init__(self):
        super().__init__("The specified address is invalid.")


class BTAdapterNotFoundError(Exception):
    """
    No Bluetooth adapter was found.
    """

    def __init__(self):
        super().__init__("No Bluetooth adapter was found.")


class DeviceNotFoundError(Exception):
    """
    The device could not be found.
    """

    def __init__(self):
        super().__init__("The device could not be found.")


class ContactingDeviceError(Exception):
    """
    The computer lost communication with the device.
    """

    def __init__(self):
        super().__init__("The computer lost communication with the device.")


class PortCouldNotBeOpenedError(Exception):
    """
    The communication port does not exist or it is already being used.
    """

    def __init__(self):
        super().__init__(
            "The communication port does not exist or it is already being used."
        )


class PortInitializationError(Exception):
    """
    The communication port could not be initialized.
    """

    def __init__(self):
        super().__init__("The communication port could not be initialized.")


class DeviceNotIdleError(Exception):
    """
    The device is not idle.
    """

    def __init__(self):
        super().__init__("The device is not idle.")


class DeviceNotInAcquisitionError(Exception):
    """
    The device is not in acquisition mode.
    """

    def __init__(self):
        super().__init__("The device is not in acquisition mode.")


class InvalidParameterError(Exception):
    """
    Invalid parameter.
    """

    def __init__(self):
        super().__init__("Invalid parameter.")


class NotSupportedError(Exception):
    """
    Operation not supported by the device.
    """

    def __init__(self):
        super().__init__("Operation not supported by the device.")


class UnknownError(Exception):
    """
    Unknown error: `message`.
    """

    def __init__(self, message=""):
        super().__init__("Unknown error: {}".format(message))



class Frame:
    """
    ScientISST Device Frame class

    A frame returned by ScientISST.read()

    Attributes:
        seq (int): Frame sequence number (0...15).

            This number is incremented by 1 on each consecutive frame, and it overflows to 0 after 15 (it is a 4-bit number).

            This number can be used to detect if frames were dropped while transmitting data.

        digital (list): Array of digital ports states (False for low level or True for high level).

            On original ScientISST, the array contents are: I1 I2 I3 I4.

            On ScientISST 2, the array contents are: I1 I2 O1 O2.

        a (list): Array of raw analog inputs values of the active channles.

            If all channels are active, `a` will have 8 elements: 6 AIs and 2 AXs.

        mv (list): Array of analog inputs values of the active channles in mV.

            If all channels are active, `mv` will have 8 elements: 6 AIs and 2 AXs.
    """

    digital = [0] * 4
    seq = -1

    def __init__(self, num_channels):
        self.a = [0] * num_channels
        self.mv = [0] * num_channels

    def to_map(self):
        return {
            "sequence": self.seq,
            "analog": self.a,
            "digital": self.digital,
            "mv": self.mv,
        }

    def __str__(self):
        if self.mv[0]:
            values = [str(val) for pair in zip(self.a, self.mv) for val in pair]
        else:
            values = map(str, self.a)

        return "{}\t{}\t{}\t{}\t{}\t{}".format(
            self.seq,
            self.digital[0],
            self.digital[1],
            self.digital[2],
            self.digital[3],
            "\t".join(values),
        )

    def to_matrix(self):
        if self.mv[0]:
            return (
                [self.seq]
                + self.digital
                + [val for pair in zip(self.a, self.mv) for val in pair]
            )
        else:
            return [self.seq] + self.digital + self.a



import sys


# if sys.platform == "linux":
import socket

# else:
import serial

import time
import re
from math import log2
import numpy as np

from scientisst.frame import *
from scientisst.state import *
from scientisst.exceptions import *
from scientisst.esp_adc.esp_adc import *
from scientisst.constants import *


class ScientISST:
    """ScientISST Device class

    Attributes:
        address (str): The device serial port address ("/dev/example") or TCP port

        serial_speed (int, optional): The serial port bitrate.
    """

    __serial = None
    __socket = None
    __num_chs = 0
    __api_mode = 1
    __sample_rate = None
    __chs = [None] * 8
    __log = False

    def __init__(
        self,
        address,
        serial_speed=115200,
        log=False,
        api=API_MODE_SCIENTISST,
        connection_tries=5,
        com_mode=COM_MODE_BT,
    ):
        """
        Args:
            address (str): The device serial port address ("/dev/example")
            serial_speed (int, optional): The serial port bitrate in bit/s
            log (bool, optional): If the bytes sent and received should be showed
            api (int): The desired API mode for the device
        """

        if (
            api != API_MODE_SCIENTISST
            and api != API_MODE_JSON
            and api != API_MODE_BITALINO
        ):
            raise InvalidParameterError()

        self.com_mode = com_mode
        self.address = address
        self.serial_speed = serial_speed
        self.__log = log

        # Setup socket in function of com_mode argument
        self.__setupSocket()

        # try to connect to board
        while True:
            try:
                # Set API mode
                self.__changeAPI(api)
                # get device version string and adc characteristics
                self.version_and_adc_chars()
                break
            except ContactingDeviceError:
                if connection_tries > 0:
                    connection_tries -= 1
                else:
                    raise ContactingDeviceError()

        sys.stdout.write("Connected!\n")

    def version_and_adc_chars(self, print=True):
        """
        Gets the device firmware version string and esp_adc_characteristics

        Returns:
            version (str): Firmware version

        Raises:
            ContactingDeviceError: If there is an error contacting the device.
        """
        if self.__api_mode == API_MODE_BITALINO:
            header = "BITalino"
        else:
            header = "ScientISST"
        header_len = len(header)

        cmd = b"\x07"
        self.__send(cmd)

        result = self.__recv(1024, waitall_flag=False)

        if result == b"":
            raise ContactingDeviceError()

        index = result.index(b"\x00")
        version = result[header_len : index - 1].decode("utf-8")

        self.__adc1_chars = EspAdcCalChars(result[index + 1 :])

        if print:
            sys.stdout.write("ScientISST version: {}\n".format(version))
            sys.stdout.write(
                "ScientISST Board Vref: {}\n".format(self.__adc1_chars.vref)
            )
            sys.stdout.write(
                "ScientISST Board ADC Attenuation Mode: {}\n".format(
                    self.__adc1_chars.atten
                )
            )

        return version

    def start(
        self,
        sample_rate,
        channels,
        reads_per_second=5,
        simulated=False,
    ):
        """
        Starts a signal acquisition from the device

        Args:
            sample_rate (int): Sampling rate in Hz.

                Accepted values are 1, 10, 100 or 1000 Hz.

            channels (list): Set of channels to acquire.

                Accepted channels are 1...6 for inputs A1...A6.

            reads_per_second (int): Number of times to read the data streaming from the device.

                Accepted values are integers greater than 0.


            simulated (bool): If true, start in simulated mode.

                Otherwise start in live mode. Default is to start in live mode.

        Raises:
            DeviceNotIdleError: If the device is already in acquisition mode.
            InvalidParameterError: If no valid API value is chosen or an incorrect array of channels is provided.
        """
        assert int(reads_per_second) > 0

        if self.__num_chs != 0:
            raise DeviceNotIdleError()

        if not channels:  # channels is empty
            chMask = 0xFF  #  all 8 analog channels
            self.__num_chs = 8
        else:
            chMask = 0
            for ch in channels:
                if ch <= 0 or ch > 8:
                    raise InvalidParameterError()
                self.__chs[self.__num_chs] = ch  # Fill chs vector

                mask = 1 << (ch - 1)
                if chMask & mask:
                    self.__num_chs = 0
                    raise InvalidParameterError()

                chMask |= mask
                self.__num_chs += 1

        self.__sample_rate = sample_rate

        # Sample rate
        sr = 0b01000011
        sr |= self.__sample_rate << 8
        self.__send(sr, 4)

        # Cleanup existing data in bluetooth socket
        self.__clear()

        if simulated:
            cmd = 0x02
        else:
            cmd = 0x01
        cmd |= chMask << 8

        self.__send(cmd)

        self.__packet_size = self.__getPacketSize()

        self.__bytes_to_read = self.__packet_size * max(
            sample_rate // reads_per_second, 1
        )
        if self.__bytes_to_read > MAX_BUFFER_SIZE:
            self.__bytes_to_read = MAX_BUFFER_SIZE - (
                MAX_BUFFER_SIZE % self.__packet_size
            )

        if self.__bytes_to_read % self.__packet_size:
            self.__num_chs = 0
            sys.stderr.write(
                "Error, bytes_to_read needs to be devisible by packet_size\n"
            )
            raise InvalidParameterError()
        else:
            self.__num_frames = self.__bytes_to_read // self.__packet_size

    def read(self, convert=True, matrix=False):
        """
        Reads acquisition frames from the device.

        This method returns when all requested frames are received from the device, or when a timeout occurs.

        Args:
            convert (bool): Convert from raw to mV
            matrix (bool): Return `Frames` in a `np.array` (matrix) form

        Returns:
            frames (list): List of [`Frame`][scientisst.frame.Frame] objects retrieved from the device. If `matrix` is True, the `frames` corresponds to a `np.array` (matrix).

        Raises:
            ContactingDeviceError: If there is an error contacting the device.
            DeviceNotInAcquisitionError: If the device is not in acquisition mode.
            NotSupportedError: If the device API is in BITALINO mode
            UnknownError: If the device stopped sending frames for some unknown reason.
        """

        frames = []

        if self.__num_chs == 0:
            raise DeviceNotInAcquisitionError()

        result = list(self.__recv(self.__bytes_to_read))
        start = 0
        for it in range(self.__num_frames):
            bf = result[start : start + self.__packet_size]
            mid_frame_flag = 0

            #  if CRC check failed, try to resynchronize with the next valid frame
            while not self.__checkCRC4(bf, self.__packet_size):
                sys.stderr.write("Error checking CRC4")
                #  checking with one new byte at a time
                result_tmp = list(self.__recv(1))
                if len(result_tmp) != 1:
                    raise ContactingDeviceError()

                result += result_tmp
                start += 1
                bf = result[start : start + self.__packet_size]

            f = Frame(self.__num_chs)
            frames.append(f)
            if self.__api_mode == API_MODE_SCIENTISST:
                # Get seq number and IO states
                f.seq = bf[-1] >> 4
                for i in range(4):
                    f.digital[i] = 0 if (bf[-2] & (0x80 >> i)) == 0 else 1

                # Get channel values
                byte_it = 0
                for i in range(self.__num_chs):
                    index = self.__num_chs - 1 - i
                    curr_ch = self.__chs[index]

                    # If it's an AX channel
                    if curr_ch == AX1 or curr_ch == AX2:
                        f.a[index] = (
                            int.from_bytes(
                                bf[byte_it : byte_it + 4], byteorder="little"
                            )
                            & 0xFFFFFF
                        )
                        byte_it += 3

                    # If it's an AI channel
                    else:
                        if not mid_frame_flag:
                            f.a[index] = (
                                int.from_bytes(
                                    bf[byte_it : byte_it + 2], byteorder="little"
                                )
                                & 0xFFF
                            )
                            byte_it += 1
                            mid_frame_flag = 1
                        else:
                            f.a[index] = (
                                int.from_bytes(
                                    bf[byte_it : byte_it + 2], byteorder="little"
                                )
                                >> 4
                            )
                            byte_it += 2
                            mid_frame_flag = 0
                        if convert:
                            f.mv[index] = self.__adc1_chars.esp_adc_cal_raw_to_voltage(
                                f.a[index]
                            )
            elif self.__api_mode == API_MODE_JSON:
                print(bf)
            else:
                raise NotSupportedError()

            start += self.__packet_size

        if len(frames) == self.__num_frames:
            if not matrix:
                return frames
            else:
                return np.array([frame.to_matrix() for frame in frames])
        else:
            raise ContactingDeviceError()

    def stop(self):
        """
        Stops a signal acquisition.

        Raises:
            DeviceNotInAcquisitionError: If the device is not in acquisition mode.
        """
        if self.__num_chs == 0:
            raise DeviceNotInAcquisitionError()

        cmd = b"\x00"
        self.__send(cmd)  # 0  0  0  0  0  0  0  0 - Go to idle mode

        self.__num_chs = 0
        self.__sample_rate = 0

        # Cleanup existing data in bluetooth socket
        self.__clear()

    def battery(self, value=0):
        """
        Sets the battery voltage threshold for the low-battery LED.

        Args:
            value (int): Battery voltage threshold. Default value is 0.

                Value | Voltage Threshold
                ----- | -----------------
                    0 |   3.4 V
                 ...  |   ...
                   63 |   3.8 V

        Raises:
            DeviceNotIdleError: If the device is in acquisition mode.
            InvalidParameterError: If an invalid battery threshold value is given.
        """
        if self.__num_chs != 0:
            raise DeviceNotIdleError()

        if value < 0 or value > 63:
            raise InvalidParameterError()

        cmd = value << 2
        # <bat threshold> 0 0 - Set battery threshold
        self.__send(cmd)

    def trigger(self, digital_output):
        """
        Assigns the digital outputs states.

        Args:
            digital_output (list): Vector of booleans to assign to digital outputs, starting at first output (O1).

        Raises:
            InvalidParameterError: If the length of the digital_output array is different from 2.
        """
        length = len(digital_output)

        if length != 2:
            raise InvalidParameterError()

        cmd = 0xB3  # 1  0  1  1  O2 O1 1  1 - Set digital outputs

        for i in range(length):
            if digital_output[i]:
                cmd |= 0b100 << i

        self.__send(cmd)

    def dac(self, voltage):
        """
        Assigns the analog (DAC) output value (ScientISST 2 only).

        Args:
            voltage (float): Analog output value to set (0V-3.3V).

        Raises:
            InvalidParameterError: If the voltage value is outside of its range, 0-255.
        """
        if voltage < 0 or voltage > 3.3:
            raise InvalidParameterError()

        cmd = 0xA3  # 1  0  1  0  0  0  1  1 - Set dac output

        # Convert from voltage to raw:
        raw = int(voltage * 255 / 3.3)

        cmd |= raw << 8
        self.__send(cmd, nrOfBytes=2)

    # TODO: test with ScientISST Sense v2
    def state(self):
        """
        Returns current device state (%ScientISST 2 only).

        Returns:
            state (State): Current device [`State`][scientisst.state.State]

        Raises:
            DeviceNotIdleError: If the device is in acquisition mode.
            ContactingDeviceError: If there is an error contacting the device.
        """
        if self.__num_chs != 0:
            raise DeviceNotIdleError()

        cmd = 0x0B
        self.__send(cmd)
        # 0  0  0  0  1  0  1  1 - Send device status

        # if (recv(&statex, sizeof statex) != sizeof statex)    # a timeout has occurred
        # throw Exception(Exception::CONTACTING_DEVICE);
        result = self.__recv(16)
        if not result or not self.__checkCRC4(result, 16):
            raise ContactingDeviceError()

        state = State()
        print(result)

        # for(int i = 0; i < 6; i++)
        # state.analog[i] = statex.analog[i];

        # state.battery = statex.battery;
        # state.batThreshold = statex.batThreshold;

        # for(int i = 0; i < 4; i++)
        # state.digital[i] = ((statex.portsCRC & (0x80 >> i)) != 0);

        # return state;

    def disconnect(self):
        """
        Disconnects from a ScientISST device. If an aquisition is running, it is stopped
        """
        if self.__num_chs != 0:
            self.stop()
        if self.__socket:
            self.__socket.shutdown(socket.SHUT_RDWR)
            self.__socket.close()
            self.__socket = None
        elif self.__serial:
            self.__serial.close()
            self.__serial = None
        sys.stdout.write("Disconnected\n")

    def __setupSocket(self):
        """
        Create a socket in function of the comunication mode desired
        """
        if self.com_mode == COM_MODE_BT:
            sys.stdout.write("Connecting to {}...\n".format(self.address))
            # Create the client socket
            if sys.platform == "linux":
                # Check if address is a valid bt MAC address
                if not re.match(
                    "[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                    self.address.lower(),
                ):
                    raise InvalidAddressError()

                self.__socket = socket.socket(
                    socket.AF_BLUETOOTH, socket.SOCK_STREAM, socket.BTPROTO_RFCOMM
                )
                self.__socket.connect((self.address, 1))
            else:
                self.__serial = serial.Serial(
                    self.address, self.serial_speed, timeout=TIMEOUT_IN_SECONDS
                )
        elif self.com_mode == COM_MODE_TCP_SERVER:
            if not self.address.isdigit():
                raise InvalidAddressError()

            port = int(self.address)

            with socket.socket() as s:
                s.bind(("", port))
                print("Binded port %d on all interfaces" % (port))

                s.listen(5)
                print("TCP Server created. Waiting for ScientISST to connect...")

                self.__socket, addr = s.accept()
                print("ScientISST with address", addr, " connected")

        elif self.com_mode == COM_MODE_TCP_AP:
            if isinstance(self.address, str):
                if not self.address.isdigit():
                    raise InvalidAddressError()
                port = int(self.address)
            elif isinstance(self.address, int):
                port = self.address
            else:
                raise InvalidAddressError()

            self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__socket.connect(("scientisst.local", port))

        else:
            raise InvalidParameterError

        self.__socket.settimeout(TIMEOUT_IN_SECONDS)

    def __getPacketSize(self):
        packet_size = 0

        if self.__api_mode == API_MODE_SCIENTISST:
            num_intern_active_chs = 0
            num_extern_active_chs = 0

            for ch in self.__chs:
                if ch:
                    # Add 24bit channel's contributuion to packet size
                    if ch == AX1 or ch == AX2:
                        num_extern_active_chs += 1
                    # Count 12bit channels
                    else:
                        num_intern_active_chs += 1

            # Add 24bit channel's contributuion to packet size
            packet_size = 3 * num_extern_active_chs

            # Add 12bit channel's contributuion to packet size
            if not (num_intern_active_chs % 2):  # If it's an even number
                packet_size += (num_intern_active_chs * 12) / 8
            else:
                packet_size += (
                    (num_intern_active_chs * 12) - 4
                ) / 8  # -4 because 4 bits can go in the I/0 byte
            # for the I/Os and seq+crc bytes
            packet_size += 2

        else:
            raise NotSupportedError()

        return int(packet_size)

    def __changeAPI(self, api):
        if self.__num_chs and self.__num_chs != 0:
            raise DeviceNotIdleError()

        if api <= 0 or api > 3:
            raise InvalidParameterError()

        self.__api_mode = api

        api <<= 4
        api |= 0b11

        self.__send(api)

    def __checkCRC4(self, data, length):
        CRC4tab = [0, 3, 6, 5, 12, 15, 10, 9, 11, 8, 13, 14, 7, 4, 1, 2]
        crc = 0
        for i in range(length - 1):
            b = data[i]
            crc = CRC4tab[crc] ^ (b >> 4)
            crc = CRC4tab[crc] ^ (b & 0x0F)

        # CRC for last byte
        crc = CRC4tab[crc] ^ (data[-1] >> 4)
        crc = CRC4tab[crc]

        return crc == (data[-1] & 0x0F)

    def __send(self, command, nrOfBytes=0):
        """
        Send data
        """

        if nrOfBytes <= 4:
            nrOfBytes = 4
        else:
            raise ValueError("Maximum send command size is 4 bytes")

        if type(command) is int:
            if command != 0:
                command = command.to_bytes(
                    int(log2(command) // 8 + 1), byteorder="little"
                )
            else:
                command = b"\x00"
        if nrOfBytes and len(command) < nrOfBytes:
            for _ in range(nrOfBytes - len(command)):
                command += b"\x00"
        # if self.__serial:
        time.sleep(0.250)
        if self.__log:
            sys.stdout.write(
                "{} bytes sent: {}\n".format(
                    len(command), " ".join("{:02x}".format(c) for c in command)
                )
            )
        if self.__socket:
            self.__socket.send(command)
        elif self.__serial:
            self.__serial.write(command)
        else:
            raise InvalidParameterError()
        # else:
        # raise ContactingDeviceError()

    def __recv(self, nrOfBytes, waitall_flag=True):
        """
        Receive data
        """
        result = None
        if self.__socket:
            if waitall_flag:
                result = self.__socket.recv(nrOfBytes, socket.MSG_WAITALL)
            else:
                result = self.__socket.recv(nrOfBytes)
        elif self.__serial:
            result = self.__serial.read(nrOfBytes)
        else:
            raise InvalidParameterError()
        if self.__log:
            if nrOfBytes > 1:
                sys.stdout.write(
                    "{} bytes received: {}\n".format(
                        nrOfBytes, " ".join("{:02x}".format(c) for c in result)
                    )
                )
            else:
                sys.stdout.write("{} bytes received: {}\n".format(1, result.hex()))
        return result

    def __clear(self):
        """
        Clear the device buffer
        """
        if self.__socket:
            self.__socket.setblocking(False)
        elif self.__serial:
            self.__serial.timeout = 0
        else:
            raise InvalidParameterError()

        try:
            while self.__recv(1):
                pass
        except BlockingIOError:
            pass

        if self.__socket:
            self.__socket.setblocking(True)
        elif self.__serial:
            self.__serial.timeout = TIMEOUT_IN_SECONDS
        else:
            raise InvalidParameterError()



TIMEOUT_IN_SECONDS = 5

# API_MODE
API_MODE_BITALINO = 1
API_MODE_SCIENTISST = 2
API_MODE_JSON = 3

# COM_MODE
COM_MODE_BT = "bt_classic"
COM_MODE_TCP_SERVER = "tcp"
COM_MODE_TCP_AP = "tcp_ap"
COM_MODE_LIST = [COM_MODE_BT, COM_MODE_TCP_SERVER, COM_MODE_TCP_AP]

# CHANNELS
AI1 = 1
AI2 = 2
AI3 = 3
AI4 = 4
AI5 = 5
AI6 = 6
AX1 = 7
AX2 = 8

MAX_BUFFER_SIZE = 4096



LIN_COEFF_A_SCALE = 65536
LIN_COEFF_A_ROUND = LIN_COEFF_A_SCALE / 2

LUT_VREF_LOW = 1000
LUT_VREF_HIGH = 1200
LUT_ADC_STEP_SIZE = 64
LUT_POINTS = 20
LUT_LOW_THRESH = 2880
LUT_HIGH_THRESH = LUT_LOW_THRESH + LUT_ADC_STEP_SIZE
ADC_12_BIT_RES = 4096

ADC_UNIT_1 = 1  # SAR ADC 1.
ADC_UNIT_2 = 2  # SAR ADC 2.
ADC_UNIT_BOTH = 3  # SAR ADC 1 and 2.
ADC_UNIT_ALTER = 7  # SAR ADC 1 and 2 alternative mode.

ADC_ATTEN_DB_0 = 0  # No input attenumation, ADC can measure up to approx. 800 mV. */
ADC_ATTEN_DB_2_5 = 1  # The input voltage of ADC will be attenuated, extending the range of measurement to up to approx. 1100 mV.
ADC_ATTEN_DB_6 = 2  # The input voltage of ADC will be attenuated, extending the range of measurement to up to  approx. 1350 mV.
ADC_ATTEN_DB_11 = 3  # The input voltage of ADC will be attenuated, extending the range of measurement to up to  approx. 2600 mV.

ADC_WIDTH_BIT_9 = 0  # ADC capture width is 9Bit. Only ESP32 is supported.
ADC_WIDTH_BIT_10 = 1  # ADC capture width is 10Bit. Only ESP32 is supported.
ADC_WIDTH_BIT_11 = 2  # ADC capture width is 11Bit. Only ESP32 is supported.
ADC_WIDTH_BIT_12 = 3  # ADC capture width is 12Bit. Only ESP32 is supported.

VOLT_DIVIDER_FACTOR = 3.399



from threading import Thread, Event
from queue import Queue
import time


class ThreadBuilder:
    def __init__(self):
        self.buffer = Queue()
        self.event = Event()
        self.thread = Thread(target=self.target)

    def start(self):
        self.thread.start()

    def put(self, frames):
        self.buffer.put(frames)

    def stop(self):
        # let the thread method finish before stop
        time.sleep(0.25)
        self.event.set()

    def target(self):
        while not self.event.is_set():
            if not self.buffer.empty():
                frames = self.buffer.get()
                self.thread_method(frames)
            else:
                time.sleep(0.1)

    def thread_method(self, frames):
        pass



import os
import sys
import importlib.util
from sense_src.thread_builder import ThreadBuilder


class CustomScript(ThreadBuilder):
    def __init__(self):
        super().__init__()
        self.on_init()

    def start(self):
        super().start()
        self.on_start()

    def stop(self):
        super().stop()
        self.on_stop()

    def thread_method(self, frames):
        self.on_read(frames)

    def on_init(self):
        pass

    def on_start(self):
        pass

    def on_stop(self):
        pass

    def on_read(self, frames):
        pass


def get_custom_script(file_path):
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    if "_" in module_name:
        module_name = "".join([word.capitalize() for word in module_name.split("_")])
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    try:
        CustomScript = getattr(module, module_name)
        return CustomScript()
    except AttributeError:
        print("%s: No such class." % module_name, file=sys.stderr)
        exit(1)



from scientisst.esp_adc.constants import *
from scientisst.esp_adc.lut_adc import *


class EspAdcCalChars:
    def __init__(self, buffer):
        self.adc_num = int.from_bytes(buffer[0:4], "little")
        self.atten = int.from_bytes(buffer[4:8], "little")
        self.bit_width = int.from_bytes(buffer[8:12], "little")
        self.coeff_a = int.from_bytes(buffer[12:16], "little")
        self.coeff_b = int.from_bytes(buffer[16:20], "little")
        self.vref = int.from_bytes(buffer[20:24], "little")

        # Initialize fields for lookup table if necessary
        if self.atten == ADC_ATTEN_DB_11:
            if self.adc_num == ADC_UNIT_1:
                self.low_curve = LUT_ADC1_LOW
            else:
                self.low_curve = LUT_ADC2_LOW

            if self.adc_num == ADC_UNIT_1:
                self.high_curve = LUT_ADC1_HIGH
            else:
                self.high_curve = LUT_ADC2_HIGH
        else:
            self.low_curve = 0
            self.high_curve = 0

    def esp_adc_cal_raw_to_voltage(self, adc_reading):
        adc_reading = adc_reading << (ADC_WIDTH_BIT_12 - self.bit_width)
        if adc_reading > ADC_12_BIT_RES - 1:
            adc_reading = ADC_12_BIT_RES - 1

        if self.atten == ADC_ATTEN_DB_11 and adc_reading >= LUT_LOW_THRESH:
            #
            lut_voltage = EspAdcCalChars.calculate_voltage_lut(
                adc_reading, self.vref, self.low_curve, self.high_curve
            )

            if adc_reading <= LUT_HIGH_THRESH:
                linear_voltage = EspAdcCalChars.calculate_voltage_linear(
                    adc_reading, self.coeff_a, self.coeff_b
                )

                voltage = EspAdcCalChars.interpolate_two_points(
                    linear_voltage,
                    lut_voltage,
                    LUT_ADC_STEP_SIZE,
                    (adc_reading - LUT_LOW_THRESH),
                )
            else:
                voltage = lut_voltage
        else:
            voltage = EspAdcCalChars.calculate_voltage_linear(
                adc_reading, self.coeff_a, self.coeff_b
            )

        return int(voltage * VOLT_DIVIDER_FACTOR)

    def interpolate_two_points(y1, y2, x_step, x):
        # Interpolate between two points (x1,y1) (x2,y2) between 'lower' and 'upper' separated by 'step'
        return int(((y1 * x_step) + (y2 * x) - (y1 * x) + int(x_step / 2)) / x_step)

    def calculate_voltage_linear(adc_reading, coeff_a, coeff_b):
        # Where voltage = coeff_a * adc_reading + coeff_b
        return (
            int(((coeff_a * adc_reading) + LIN_COEFF_A_ROUND) / LIN_COEFF_A_SCALE)
            + coeff_b
        )

    def calculate_voltage_lut(adc, vref, low_vref_curve, high_vref_curve):

        # Get index of lower bound points of LUT
        i = int((adc - LUT_LOW_THRESH) / LUT_ADC_STEP_SIZE)

        # Let the X Axis be Vref, Y axis be ADC reading, and Z be voltage
        x2dist = LUT_VREF_HIGH - vref  # (x2 - x)
        x1dist = vref - LUT_VREF_LOW  # (x - x1)
        y2dist = ((i + 1) * LUT_ADC_STEP_SIZE) + LUT_LOW_THRESH - adc  # (y2 - y)
        y1dist = adc - ((i * LUT_ADC_STEP_SIZE) + LUT_LOW_THRESH)  # (y - y1)

        # For points for bilinear interpolation
        q11 = low_vref_curve[i]  # Lower bound point of low_vref_curve
        q12 = low_vref_curve[i + 1]  # Upper bound point of low_vref_curve
        q21 = high_vref_curve[i]  # Lower bound point of high_vref_curve
        q22 = high_vref_curve[i + 1]  # Upper bound point of high_vref_curve

        # Bilinear interpolation
        # Where z = 1/((x2-x1)*(y2-y1)) * ( (q11*x2dist*y2dist) + (q21*x1dist*y2dist) + (q12*x2dist*y1dist) + (q22*x1dist*y1dist) )
        voltage = (
            (q11 * x2dist * y2dist)
            + (q21 * x1dist * y2dist)
            + (q12 * x2dist * y1dist)
            + (q22 * x1dist * y1dist)
        )
        voltage += int(((LUT_VREF_HIGH - LUT_VREF_LOW) * LUT_ADC_STEP_SIZE) / 2)
        # Integer division rounding
        voltage = int(voltage / (LUT_VREF_HIGH - LUT_VREF_LOW) * LUT_ADC_STEP_SIZE)
        # Divide by ((x2-x1)*(y2-y1))

        return voltage



import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from scientisst.constants import *


class ArgParser:
    class MyParser(ArgumentParser):
        def error(self, message):
            sys.stderr.write("error: %s\n\n" % message)
            self.print_help()
            sys.exit(2)

    def __init__(self):
        usage = "%(prog)s [args] address"
        description = "description: The program connects to the ScientISST Sense device and starts an acquisition, providing the option to store the received data in a .csv file."
        self.parser = self.MyParser(
            usage=usage, description=description, formatter_class=RawTextHelpFormatter
        )

        self.parser.add_argument(
            "address",
            nargs="?",
            type=str,
            help="For BTH communication:\n\tLinux: BTH MAC address\n\tMac: serial port address\n\tWindows: BTH serial COM port\nFor TCP/UDP communication:\n\tAll plataforms: server port.",
        )
        self.parser.add_argument(
            "-f",
            "--frequency",
            dest="fs",
            help="sampling frequency, default: 1000",
            type=int,
            default=1000,
        )
        self.parser.add_argument(
            "-c",
            "--channels",
            dest="channels",
            type=str,
            help="analog channels, default: 1,2,3,4,5,6",
            default="1,2,3,4,5,6",
        )
        self.parser.add_argument(
            "-d",
            "--duration",
            dest="duration",
            help="duration in seconds, default: unlimited",
            type=int,
            default=0,
        )
        self.parser.add_argument(
            "-o",
            "--output",
            dest="output",
            help="write report to output file, default: None",
            type=str,
            default=None,
        )
        self.parser.add_argument(
            "-r",
            "--raw",
            action="store_false",
            dest="convert",
            default=True,
            help="do not convert from raw to mV",
        )
        self.parser.add_argument(
            "-s",
            "--lsl",
            dest="stream",
            action="store_true",
            default=False,
            help="stream data using Lab Streaming Layer protocol. Use `python -m pylsl.examples.ReceiveAndPlot` to view stream",
        )
        self.parser.add_argument(
            "--script",
            dest="script",
            help="send the received frames to a script that inherits the CustomScript class",
            type=str,
            default=None,
        )
        self.parser.add_argument(
            "-q",
            "--quiet",
            action="store_false",
            dest="verbose",
            default=True,
            help="don't print ScientISST frames",
        )
        self.parser.add_argument(
            "-v",
            "--version",
            dest="version",
            action="store_true",
            default=False,
            help="show sense.py version",
        )
        self.parser.add_argument(
            "--verbose",
            dest="log",
            action="store_true",
            default=False,
            help="log sent/received bytes",
        )
        self.parser.add_argument(
            "-m",
            "--mode",
            dest="mode",
            type=str,
            default=COM_MODE_BT,
            help="The communication mode. Currently supported modes: "
            + ", ".join(COM_MODE_LIST)
            + ". Default: "
            + COM_MODE_BT,
        )
        self.args = self.parser.parse_args()

    def error(self, value):
        self.parser.error(value)
